%% CombinedSolver_inner
%   b   : [Bool]  Bitmask of sensors. True means the Sensor is not active, or rather "sensor is under attack"
%   cc  : [CC]    Convex constraints (reals implicitly <= 0)
%   learnings : [[Int]] Learned clauses as indicies into [b c].
%   s   : Int     Security bound: \forall i . sum b_i <= s
%
% Pre: size(b) == size(s)
% Post: \forall i ccVars Assignment.b(i) and Assignment.cc(ccVars) satisfy the
% combined boolean convex problem.
%
% Design decision: We could keep a 'phi' variable that accumulates the initial
% problem and all clauses learned from our certificates, but by reifying we can
% optionally filter the learnings and perform garbage collection.

function [unsatSATConflict Assignment totalTheoryTime] = CombinedSolver_inner(b,cc,learnings,s)
  [nr nr] = size(cc);
  totalTheoryTime = 0;

  % 1) Build a pseudo boolean version of the problem
  %    by associating a boolean to each convex constraint, cc.
  phi = Phi(b,learnings,s);

  'Computing a satisfying boolean assignment for attack vector'
  [isSat bModel] = check_sat(phi);

  if(isSat)
    % 3) Get the active clauses.
    [ccActive bActive] = getActive(bModel.b,cc,b);

    % 4) Run the theory solver, if ccActive is SAT return the model otherwise return the conflict.
    'Boolean SAT.  Consulting theory solver'
    start = clock;
    [tSat tModel] = theorySolver(ccActive);
    finish = etime(clock,start);
    totalTheoryTime = totalTheoryTime + finish;
    if(tSat)
        'Reals SAT.  Done.'
        unsatSATConflict = 1; % SAT
        Assignment = buildModel(bModel,tModel);
    else
        'Reals UNSAT'
        unsatSATConflict = 2; % conflict
        start = clock;
        Assignment = [learnings (unsatCore(ccActive,bActive))];
        finish = etime(clock,start);
        totalTheoryTime = totalTheoryTime + finish;
    end
  else
      'Booleans UNSAT'
      unsatSATConflict = 0; % UNSAT.
      Assignment = [];
  end
end

function model = buildModel(boolModel, realModel)
    realModel.b = boolModel.b;
    model       = realModel;
end

% Given the sensor mask, mask inverse, past learned sat-cores, and security
% parameter (min popcnt for the mask), build a SAT problem to determine if a valid
% satisfying assignment exists.
function phi = Phi(b,learnings,s)
    [nr nr] = size(b);
    phi = true;
    sm  = 0;
    % forall i . not b_i => c(i)
    for i = [1:nr]
        % by def of 'c': phi = phi && (b(i) || c(i));
        sm = sm + b(i);
    end
    % The number of discarded constraints is bounded by s
    phi = phi && (sm <= s);

    [nrL nrL] = size(learnings);
    % Add previously learned clauses to the formulation
    for i = [1:nrL]
       phi = phi && learnings(i);
   end
end

% Check if the constraints are feasible
% We do this by providing a slack variable
% to each constraint then try to optimize,
% thus obtaining sum(slacks)=0 for feasible
% constraints.
function [tSat,tModel] = theorySolver(cc)
    [nr nr] = size(cc);
    slack   = symbolic('theorySlack', 'real');

    % Optimize 'slack' subject to slack >=0, cc(i) - slack <= 0
    constrs = [-slack];
    for i = [1:nr]
        constrs = [constrs (cc(i) - slack)]; % constraints are implicilty <= 0
    end
    [tFeasible tModel] = convex_opt(slack, constrs);

    % If slack ~ 0 then this is _probably_ feasible... need an eval() function
    % to be sure (future work).
    if tFeasible
        tSat = tModel.theorySlack < 0.0001;
    else
        tSat = false;
    end
end

% (\forall i . c(i) => cc(i) \in ccActive(i))
function [ccActive bActive] = getActive(bModel,cc,b)
    [nr nr] = size(bModel);
    ccActive = [];
    bActive  = [];
    for i = [1:nr]
        if(not(bModel(i)))
            ccActive = [ccActive ( cc(i) )];  % Constraints are implicitly <= 0.
            bActive  = [bActive b(i)];        % For building unsat cores later,
                                              % we need the attack vector under consideration
        end
    end
end
