% This is the basic "incrementally greedy" search procedure
% for the 1D string-packing problem.

% Parameters to the problem
len        = 1000; % how long is the input stream
width      =    5; % how wide is the narrow sliding avg. window
extra      =   30; % how much wider is the wide sliding avg. window
maxsearch  =   50; % how far forward to search for a location before resetting
limit      =    1; % what is the limit for the wide sliding average?
searchinc  =    5; % how many spaces to skip when we search foraward
inputlimit =  300; % maximum distance from 0 an input value can be

% the strings we want to pack
segments  = { 'hello!asdf' ...
            , 'whatup'  ...
            , 'yall' ...
            , 'howdy', 'pardner' ...
            , 'WHATDO', 'YOUWANT' ...
            , 'howareyou' ...
            , 'my man'
            };
numsegs   = size(segments, 2);

% The symbolic input vector we are searching for
input     = symbolic('input', 1, len, 'integer');

% The vector of locations at which we place the strings
locs      = zeros(1, numsegs+1);

% symbolically calculate the narrow sliding window avg.
y = slide( input, width, 1 );

% symbolically calculate the wide sliding window avg.
z = slide( input, width+extra, 1 );


%% Now begin the location search algorithm

%  1) The outer loop gradually increases the 'minsep' value which
%     determines the minimum amount of separation between packed
%     strings.  It continues until eventually we find a solution.
minsep = 70;
c = false;
while ( not(c) )
    minsep = minsep + searchinc;

    % 2) The next loop tries to pack each string in turn
    for i = 1:numsegs
        if( i > 1 )
            start = locs(i-1) + minsep + size(segments{i-1},2);
        else
            start = minsep;
        end

        %  3) The third loop starts 'minsep' locations from the end of the
        %     previous string, trying successive locations up to
        %     'maxsearch' away
        offset = start;
        while offset <= start+maxsearch
            locs(i) = offset;
            locs(i+1) = offset + minsep + size(segments{i},2);
            offset = offset + searchinc;

            % display the current location vector
            locs

            % 4) Here we build the assertion that all the strings
            %    up to string i appear in the output, and that the
            %    correctly link up to the next string in the sequence.
            p = true;
            for j = 1:i
                segsize = size(segments{j}, 2);
                for k = 1:segsize
                    p = p & (y(locs(j) + k - 1) == segments{j}(k));
                end
                if( j < numsegs )
                    p = p & (y( locs(j) + segsize ) == (locs(j+1) - (locs(j) + segsize)));
                else
                    p = p & (y( locs(j) + segsize ) == 0 );
                end
            end

            % Here we bound the size of the larger sliding average:
            % this has the general effect of forcing the input
            % vector to be roughly centered around 0.
            p = p & all ( z <= limit ) & all ( z >= -limit );

            % We also directly bound the magnitude of the input values
            p = p & all ( input <= inputlimit ) & all (input >= -inputlimit );

            % Try to find a solution
            [c r] = check_sat( p )

            % If we succeed at this step, break out of loop 3 and
            % go on to try the next string.
            if( c )
                break;
            end
        end

        % If we ended the previous loop without a success, we
        % failed to pack the ith string.  Break out of loop 2
        % and start over with an increased 'maxsep'
        if( not(c) )
            break;
        end
    end
end

% Print the results of the narrow sliding average
q = slide( r.input, width, 1 )

% Print the results of the wide sliding average
w = slide( r.input, width+extra, 1 )

% Print the segments of the output where we expect to find
% the desired strings packed
for i = 1:numsegs
  q( locs(i) : (locs(i)+size(segments{i},2)) )
end
