% This is the CAMUS-based search procedure for the
% 1D string-packing problem.

% Parameters to the problem
len        = 1000; % how long is the input stream
width      =    5; % how wide is the narrow sliding avg. window
extra      =   30; % how much wider is the wide sliding avg. window
maxsearch  =  250; % how far forward to search for a location before resetting
limit      =    1; % what is the limit for the wide sliding average?
searchinc  =    1; % how many spaces to skip when we search foraward
inputlimit =  300; % maximum distance from 0 an input value can be

% the strings we want to pack
segments  = { 'hello!asdf' ...
            , 'whatup'  ...
            , 'yall' ...
            , 'howdy', 'pardner' ...
            , 'WHATDO', 'YOUWANT' ...
            , 'howareyou' ...
            , 'my man'
            };
numsegs   = size(segments, 2);

% The symbolic input vector we are searching for
input     = symbolic('input', 1, len, 'integer');

% The vector of locations at which we place the strings
sym_gaps   = symbolic('gaps', 1, numsegs, 'integer');

% symbolically calculate the narrow sliding window avg.
y = slide( input, width, 1 );
y_len = size(y, 2);

% symbolically calculate the wide sliding window avg.
z = slide( input, width+extra, 1 );

%% Now begin the location search algorithm

%  1) The outer loop gradually increases the 'minsep' value which
%     determines the minimum amount of separation between packed
%     strings.  It continues until eventually we find a solution.
minsep = 70;
c = false;
more_models = true;
while ( not(c) & minsep < len )
    minsep = minsep + searchinc

    sym_locs = locs_from_gaps(segments, sym_gaps);

    locs_ok = sym_locs(1) >= 50 & sym_locs(numsegs) <= y_len - 50;
    for i = 1:numsegs
        % every location is in range
        locs_ok = locs_ok & minsep <= sym_locs(i) & ...
                  sym_locs(i) + size(segments{i},2) + minsep < y_len;

        for j = 1:numsegs
            if i < j
                ij_disjoint = sym_locs(i) + size(segments{i},2) + minsep <= sym_locs(j);
                % sym_locs(j) + size(segments{j},2) + minsep <= sym_locs(i);
                locs_ok = locs_ok & ij_disjoint;
            end
        end
    end

    locmodels = 0;
    [more_models loc_model] = check_sat( locs_ok )

    while( more_models & locmodels < maxsearch )
        locmodels = locmodels + 1
        gaps = loc_model.gaps
        locs = locs_from_gaps( segments, loc_model.gaps )

        % Here we build the assertion that all the strings
        % up to string i appear in the output, and that the
        % correctly link up to the next string in the sequence.
        clauses = zeros(1, numsegs);
        for j = 1:numsegs
            clauses(j) = true;
            segsize = size(segments{j}, 2);
            for k = 1:segsize
                clauses(j) = clauses(j) & (y(locs(j) + k - 1) == segments{j}(k));
            end
            if( j < numsegs )
                clauses(j) = clauses(j) & (y( locs(j) + segsize ) == (locs(j+1) - (locs(j) + segsize)));
            else
                clauses(j) = clauses(j) & (y( locs(j) + segsize ) == 0 );
            end
        end

        % Here we bound the size of the larger sliding average:
        % this has the general effect of forcing the input
        % vector to be roughly centered around 0.
        p = all ( z <= limit ) & all ( z >= -limit );
        
        % We also directly bound the magnitude of the input values
        p = p & all ( input <= inputlimit ) & all (input >= -inputlimit );

        % Try to find a solution
        [c r] = check_sat( p & all( clauses ) )

        if( c )
            break;
        end

        %mcs = find_mcs( p, clauses )

        %cores = find_mus( p, clauses )
        cores = find_all_mus( p, clauses, 10 )
        % for i = 1:size(cores,1)
        %    verify_core( p, clauses, cores(i,:) );
        % end

        contpred = true;
        haspairs = false;
        for coreidx = 1:size(cores,1)
            core = cores(coreidx,:);
            
            haspairs = false;
            corepred = false;
            for i = 1:numsegs
                if core(i)
                    for j = i+1:numsegs
                        if core(j)
                            haspairs = true;
                            corepred = corepred | ...
                                sym_locs(j) - sym_locs(i) >= ...
                                locs(j) - locs(i) + searchinc;
                        end
                    end
                end
            end
            if not(haspairs)
                for i = 1:numsegs-1
                    if core(i)
                        corepred = sym_gaps(i+1) >= gaps(i+1) + searchinc;
                        break;
                    end
                end
                if core(numsegs)
                    corepred = sym_locs(numsegs) <= locs(numsegs) - searchinc;
                end
            end

            contpred = contpred & corepred;
        end

        % get another location vector
        locs_ok = locs_ok & contpred;
        [more_models loc_model] = check_sat( locs_ok );
    end
end

if c 
    % Print the results of the narrow sliding average
    q = slide( r.input, width, 1 )

    % Print the results of the wide sliding average
    w = slide( r.input, width+extra, 1 )

    % Print the segments of the output where we expect to find
    % the desired strings packed
    for i = 1:numsegs
        q( locs(i) : (locs(i)+size(segments{i},2)) )
    end
else
    'No solution found!!'
end
